#
# zadolbator: simple reminder component for Jabber service
#
# Copyright (C) 2007 Alexander Tsvyashchenko
#
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

#
# Based on RBIRC-T transport
#
require 'rubygems'

gem 'xmpp4r'
gem 'icalendar'
gem 'vpim'

require 'xmpp4r'
require 'icalendar'
require 'vpim'

require 'xmpp4r/version'
require 'xmpp4r/discovery'
require 'xmpp4r/iq'
require 'xmpp4r/dataforms'

# Config support
require 'yaml'

require 'date'

# DB storage support
require 'zadolbator/storage.rb'

# Commands
require 'zadolbator/common_cmds.rb'
require 'zadolbator/add_events.rb'
require 'zadolbator/search_events.rb'
require 'zadolbator/import_events.rb'
require 'zadolbator/modify_events.rb'
require 'zadolbator/remove_events.rb'
require 'zadolbator/subscribe_to_events.rb'

# Worker thread
require 'zadolbator/worker_thread.rb'

Jabber::debug = true
NS_COMMANDS = 'http://jabber.org/protocol/commands'

class ZadolbatorComponent < Jabber::Component

  attr_reader :log

  ##
  # Load configuration file from a YAML document named [configfile]
  #
  # Connect the Jabber component and add callbacks
  def initialize(config, log, log_level)
    @config = YAML::load File.new(config)
    @log = Logger.new(log, 10, 1024*1024)
    @log.level = log_level

    super jabber_conf['jid']
    connect jabber_conf['server'], jabber_conf['port']
    auth jabber_conf['password']

    Storage.init_connection(reminder_conf['db_adapter'], reminder_conf['db_name'])

    @cmd_procs = {
                   AddEventsCmd.cmd_name => AddEventsCmd.new,
                   ImportEventsCmd.cmd_name => ImportEventsCmd.new,
                   ModifyEventsCmd.cmd_name => ModifyEventsCmd.new,
                   RemoveEventsCmd.cmd_name => RemoveEventsCmd.new,
                   SubscribeToEventsCmd.cmd_name => SubscribeToEventsCmd.new
                 }

    @cmd_msg_procs = {
                       AddEventsMsgCmd.cmd_name => AddEventsMsgCmd.new,
                       RemoveEventsMsgCmd.cmd_name => RemoveEventsMsgCmd.new,
                       SearchEventsMsgCmd.cmd_name => SearchEventsMsgCmd.new,
                       ModifyEventsMsgCmd.cmd_name => ModifyEventsMsgCmd.new,
                       SubscribeToEventsMsgCmd.cmd_name => SubscribeToEventsMsgCmd.new
                     }

    add_iq_callback { |iq| handle { handle_iq(iq) } }
    add_message_callback { |msg| handle { handle_message(msg) } }
    add_presence_callback { |pres| handle { handle_presence(pres) } }

    @log.debug('Sending presense to all subscrived users')

    Storage::User.find(:all).each do |row|
      send_advert(row.id)
    end

    @log.debug('Spawning server thread')

    @worker_thread = Zadolbator::ServerThread.new(self, reminder_conf)

    Thread.new { @worker_thread.send_reminders_in_loop }
  end

  ##
  # Return the Jabber-specific part of the configuration
  def jabber_conf
    @config['jabber']
  end

  ##
  # Return the Reminder-specific part of the configuration
  def reminder_conf
    @config['reminder']
  end

  def handle(&handler)
    begin
      handler.call
    rescue => ex
      @log.fatal("Unhandled exception: #{ex}")
    end
  end

  def handle_iq(iq)
    @log.debug("#{iq.from} #{iq.queryns} to #{iq.to}")
    if iq.query.kind_of?(Jabber::Discovery::IqQueryDiscoInfo)
      handle_disco_info(iq)
      true
    elsif iq.query.kind_of?(Jabber::Discovery::IqQueryDiscoItems)
      handle_disco_items(iq)
      true
    elsif iq.first_element('command')
      handle_iq_command(iq)
    else
      false
    end
  end

  def handle_disco_info(iq)
    if (iq.type == :get)
      iq.type = :result
      iq.query = Jabber::Discovery::IqQueryDiscoInfo.new
      if iq.to.node == nil
        iq.query.add Jabber::Discovery::Identity.new('headline', jabber_conf['identity'], 'reminder')
      else
        iq.type = :error
        iq.add_element(Jabber::Error.new('item-not-found'))
      end
      [Jabber::Discovery::IqQueryDiscoInfo.new.namespace,
       Jabber::Discovery::IqQueryDiscoItems.new.namespace,
       NS_COMMANDS].each { |ns|
        iq.query.add(Jabber::Discovery::Feature.new(ns))
      }
    else
      iq.type = :error
      iq.add_element(Jabber::Error.new('bad-request'))
    end
    iq.to, iq.from = iq.from, iq.to
    send(iq)
  end

  def handle_disco_items(iq)
    if iq.type == :get && iq.query.node == NS_COMMANDS
      iq.type = :result
      populate_commands(iq.query)
    else
      iq.type = :error
      iq.add_element(Jabber::Error.new('bad-request'))
    end
    iq.to, iq.from = iq.from, iq.to
    send(iq)
  end

  def handle_iq_command(iq)
    cmd = iq.first_element('command')

    if cmd.attributes['node']
      if not @cmd_procs.has_key?(cmd.attributes['node'])
        iq.type = :error
        iq.add_element(Jabber::Error.new('bad-request'))
      else
        @worker_thread.refill_occurences() if @cmd_procs[cmd.attributes['node']].process(iq, @log)
      end
    end
    iq.to, iq.from = iq.from, iq.to
    send(iq)
  end

  def handle_message(msg)
    @log.debug("Message from #{msg.from} to #{msg.to} (#{msg.type}): #{msg.body.inspect}")
    return if msg.type == :error or not msg.body
    body = StringIO.open(msg.body)
    cmd = body.readline.chomp
    @log.debug("Received message with command #{cmd}")
    if not @cmd_msg_procs.has_key?(cmd)
     if cmd.upcase != "HELP"
       msg.body = "Unknown command: #{cmd}\n\n"
     else
       msg.body = "Help on Zadolbator message-based interface usage:\n\n"
     end
     msg.body += "Available commands:\n\n"
     @cmd_msg_procs.values.each { |val|
       msg.body += "Command: #{val.cmd_name}\n"
       msg.body += "Instructions: #{val.cmd_help}\n\n"
     }
    else
      @worker_thread.refill_occurences() if @cmd_msg_procs[cmd].process(msg, @log)
    end

    msg.to, msg.from = msg.from, msg.to
    send(msg)
  end

  def handle_presence(pres)
    @log.debug("presence: from #{pres.from} type #{pres.type} to #{pres.to}")

    # Handle subscription request
    if pres.type == :subscribe
      # Accept subscription
      send(Jabber::Presence.new.set_from(pres.to).set_to(pres.from).set_type(:subscribed))
      # Add user to DB, if needed
      if not Storage::User.exists?(pres.from.strip.to_s)
        new_user = Storage::User.new
        new_user.id = pres.from.strip.to_s
        new_user.save
      end
      send_advert(pres.from.strip)
      # Subscribe to sender
      send(Jabber::Presence.new.set_from(pres.to).set_to(pres.from).set_type(:subscribe))
    elsif pres.type == :unsubscribe
      send(Jabber::Presence.new.set_from(pres.to).set_to(pres.from).set_type(:unsubscribed))
      Storage::User.destroy(pres.from.strip.to_s)
      send(Jabber::Presence.new.set_from(pres.to).set_to(pres.from).set_type(:unsubscribe))
    elsif pres.type == nil && Storage::User.exists?(pres.from.strip.to_s)
      send_advert(pres.from.strip)
    end
  end

  def populate_commands(query)
    @cmd_procs.each do |procs|
      procs[1].add_cmd(jabber_conf['jid'], query)
    end
  end

  def send_advert(jid)
      send(Jabber::Presence.new.set_from(jabber_conf['jid']).set_to(jid))

      msg = Jabber::Message.new.set_from(jabber_conf['jid']).set_to(jid)
      msg.add(query = Jabber::IqQuery.new)
      query.add_namespace(Jabber::Discovery::IqQueryDiscoItems.new.namespace)
      query.attributes['node'] = NS_COMMANDS
      populate_commands(query)
      send(msg)
  end

  ##
  # Shutdown the component
  #
  # Instructs all clients to quit and
  # waits until no active clients left
  def shutdown!
    @log.info('Beginning shutdown')

    @log.debug('Sending presence:unavailable to all subscribed users ...')

    Storage::User.find(:all).each do |row|
      send(Jabber::Presence.new.set_from(jabber_conf['jid']).set_to(row.id).set_type(:unavailable))
    end

    @log.debug('Finished')

    close
  end
end
