#
# zadolbator: simple reminder component for Jabber service
#
# Copyright (C) 2007 Alexander Tsvyashchenko
#
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU
# General Public License as published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not,
# write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

module Zadolbator

  class ServerThread

    def initialize(comp, conf)
      @component = comp
      @occurs = []
      @next_fill_date = DateTime.now
      @current_date = @next_fill_date
      @occurs_lock = Mutex.new
    end

    def send_reminder(info)
      e = Storage::Event.find(info[:uid])
      e.users.each { |user|
        @component.log.debug("Sending reminder #{info[:uid]} to #{user.id}")
        msg = Jabber::Message.new.set_from(@component.jabber_conf['jid']).set_to(user.id)
        cals = Icalendar::parse(e.ical)
        # We know only one calendar and one event should be present
        event = cals.first.events.first

        if @component.reminder_conf['allow_ruby_script'] and (ruby_script = event.properties['x-ruby_script']) != nil
          @component.log.debug("Script: #{ruby_script}")
          begin
            safe_level = @component.reminder_conf['ruby_script_safe']
            safe_level = 3 if safe_level == nil
            ruby_res = eval "$SAFE = #{safe_level}; #{ruby_script}"
          rescue => ex
            @component.log.error(ex)
          end
        end

        msg.body = "#{event.dtstart} - #{event.summary}\n#{info[:desc]}\n#{ruby_res}\n\n\n#{e.ical}"
        @component.send(msg)
      }
    end

    # Fill occurences for all events until 'date_end'
    def fill_occurences(date_start)
      date_end = date_start + @component.reminder_conf['queue_interval'] / (24.0 * 3600)
      @occurs.clear
      Storage::Event.find(:all).each { |e|
        cals = Icalendar::parse(e.ical)
        @component.log.debug(e.ical)
        event = cals.first.events.first

        event.alarms.each { |alarm|
          # If events are repeated - put their durations into array.
          # At least first item is always filled.
          durs = [0]
          if alarm.duration?
            dur = Vpim::Icalendar::decode_duration(alarm.duration)
            alarm.repeat.times { |i| durs << (i + 1) * dur }
          end

          @component.log.debug("Durations: #{durs.inspect}")

          rel = alarm.trigger.ical_params['RELATED']
          raise "RELATED=END option is not supported yet" if rel and rel.first == 'END'

          @component.log.debug(alarm.trigger.ical_params['VALUE'])

          if alarm.trigger.ical_params['VALUE'] and alarm.trigger.ical_params['VALUE'].first == 'DATE-TIME'
            # The easy one: start date is supplied directly
            @component.log.debug("Alarm date: #{alarm.trigger}")
            durs.each { |dur|
              cur_occ = DateTime.parse(alarm.trigger) + dur / (24.0 * 3600)
              @component.log.debug("Actual occ: #{cur_occ}")
              @occurs << {:time => cur_occ, :uid => e.id, :desc => alarm.description} if date_start <= cur_occ and cur_occ <= date_end
            }
          else
            trig_dur = Vpim::Icalendar::decode_duration(alarm.trigger)
            @component.log.debug("Trigger duration: #{trig_dur}, start time: #{event.dtstart}")
        
            # Vpim does not work in DateTime (btw, why???), but uses Time instead -
            # thus, it cannot handle years before 1970. To overcome that, we replace here
            # all years before 1970 by 1970, hoping that recurrence is not more than yearly,
            # so this replacement, being done for calculations only, does not break things.
            if event.dtstart.year < 1970
              # Is there any better way to just replace a year in date???
              start = DateTime.civil(1970, event.dtstart.month(), event.dtstart.day(),
                                     event.dtstart.hour(), event.dtstart.min(), event.dtstart.sec(),
                                     event.dtstart.offset(), event.dtstart.start())
            else
              start = event.dtstart
            end

            # We know only one calendar should be present
            rrule = Vpim::Rrule.new(Time.parse(start.strftime("%c")), event.recurrence_rules.first)
            @component.log.debug("date_end: #{date_end}, durs.min: #{durs.min}, trig_dur: #{trig_dur}")
            rrule.each(Time.parse(date_end.strftime("%c")) - durs.min - trig_dur) { |occ|
              @component.log.debug("Occ: #{occ}")
              durs.each { |dur|
                cur_occ = DateTime.parse((occ + dur + trig_dur).iso8601)
                @component.log.debug("Actual occ: #{cur_occ}")
                @occurs << {:time => cur_occ, :uid => e.id, :desc => alarm.description} if date_start <= cur_occ and cur_occ <= date_end
              }
            }
          end
        }
      }
      @occurs.sort_by { |occ| occ[:time] }
      @component.log.debug("Occurences: #{@occurs.inspect}")
      @next_fill_date = date_end
    end

    def send_reminders_in_loop
      while true
        @occurs_lock.synchronize {
          now = DateTime.now
          @component.log.debug("Current time: #{now}")
          if not @occurs.empty?
            @component.log.debug("First time: #{@occurs.first[:time]}")
            while not @occurs.empty? and @occurs.first[:time] <= now
              send_reminder(@occurs.shift)
            end
          end
          fill_occurences(now) if @next_fill_date <= now
          @current_date = now
        }
        sleep(@component.reminder_conf['check_interval'])
      end
    end

    def refill_occurences()
      @occurs_lock.synchronize {
        fill_occurences(@current_date)
      }
    end
  end
end
